""" Description. """

# Module information
# ==================

from .version_info import * # noqa
from .main import Linter # noqa


__version__ = VERSION
__project__ = PROJECT
__author__ = AUTHOR
__license__ = LICENSE
